# Some custom plotting functions
from matplotlib import pyplot as plt # used for concurrent plots, todo go to plotly
import plotly.express as px


def plot_speed_comparison(thresh_time, scores_time, scores_it, thresh_it=None, legend=None, title=None, xlabel="Rec error threshold", ylabel="Number of faster runs"):
    fig0 = plt.figure()
    plt.subplot(121)
    plt.semilogx(thresh_time, scores_time.T)
    if not thresh_it:
        thresh_it = thresh_time
    if legend:
        plt.legend(legend)
    if title:
        plt.title(title)
    else:
        plt.title("Faster runs in time")
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.subplot(122)
    plt.semilogx(thresh_it, scores_it.T)
    if legend:
        plt.legend(legend)
    if title:
        plt.title(title)
    else:
        plt.title("Faster runs in iteration")
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    return fig0

def plot_conv():
    # TODO: subplots with iterations and time
    return



import pandas as pd
import numpy as np
pd.options.mode.chained_assignment = None  # default='warn'

# todolist:
# - transform df to add error at timestamps/itstamps
# - build new df for convergence plots
# - build scores for comparison plots

def find_best_at_all_thresh(df, thresh, batch_size, err_name="errors", time_name="timings"):
    """
    This utility function finds the method was the fastest to reach a given threshold, at each threshold in the list thres.

    Parameters
    -----------
    df : Pandas DataFrame
        The dataframe containing the errors and timings for each algorithm at each iterations, for several runs.
        Because I am a lazy coder:

            - Batch size must be constant
            - The algorithms must always be stored in df in the same order

    thresh: list
        A list of thresholds to be used for computing which method was faster.

    batch_size: int
        Number of algorithm runs to compare for the max pooling. Should be a multiple (typically 1x) of the number of algorithms.

    Returns
    --------
    scores_time: nd array
        A table (method x thresh) with how many times each method was the fastest to reach a given threshold. Here faster is understood in runtime.

    scores_it: nd array
        A table (method x thresh) with how many times each method was the fastest to reach a given threshold. Here faster is understood in number of iterations.
    """
    #TODO: remove batch parameter/provide clever default

    timings = []
    iterations = []
    # Strategy: we sweep each error and find at which time each threshold was attained
    for row_errs,errs in enumerate(df[err_name]):
        pos = 0
        time_reached = []
        it_reached = []
        for pos_err,err in enumerate(errs):
            while pos<len(thresh) and err<thresh[pos]:
                # just in case several thresholds are beaten in one iteration
                time_reached.append(df[time_name][row_errs][pos_err])
                it_reached.append(pos_err)
                pos+=1
        if len(time_reached)<len(thresh):
            time_reached = time_reached +( (len(thresh)-len(time_reached))*[np.Inf] )
            it_reached = it_reached +( (len(thresh)-len(it_reached))*[np.Inf] )
        timings.append(time_reached)
        iterations.append(it_reached)
    # casting as a numpy array (matrix) for slicing vertically
    timings = np.array(timings)
    iterations = np.array(iterations)

    # Then we find who is the winner for each batch and for each threshold
    Nb_batch = int(len(timings)/batch_size)  # should be integer without recast
    # reshaping timings into a 3-way tensor for broadcasting numpy argmax
    timings = np.reshape(timings, [Nb_batch,batch_size,len(thresh)])
    iterations = np.reshape(iterations, [Nb_batch,batch_size,len(thresh)])
    # we can now find which count how many times each algorithm was faster by finding the index of the fastest method for each batch
    winners_time = my_argmin(timings)
    winners_it = my_argmin(iterations)
    # Assuming results are stored always in the same order, a batch row index corresponds to an algorithm name
    scores_time = np.zeros((batch_size,len(thresh)))
    scores_it = np.zeros((batch_size,len(thresh)))
    for k in range(batch_size):
        for i in range(Nb_batch): 
            for j in range(len(thresh)):
                if type(winners_time[i,j])==list:
                    if k in winners_time[i,j]:
                        scores_time[k,j]+=1
                else:
                    if winners_time[i,j]==k:
                        scores_time[k,j]+=1
                    
                if type(winners_it[i,j])==list:
                    if k in winners_it[i,j]:
                        scores_it[k,j]+=1
                else:
                    if winners_it[i,j]==k:
                        scores_it[k,j]+=1

    return scores_time, scores_it, timings, iterations

def my_argmin(a):
    """
    argmin but returns list of equal indices. Axis must be 1, a is a third order tensor.
    """
    tutu = a.min(axis=1)[:,None]
    tutu[tutu==np.Inf]=0 #improve? 
    minpos = (a == tutu)
    npargmin = np.argmin(a,axis=1)
    myargmin= np.zeros(npargmin.shape, dtype=object)-1
    for i in range(minpos.shape[0]):
        for j in range(minpos.shape[1]):
            for k in range(minpos.shape[2]):
                if minpos[i,j,k]:
                    if type(myargmin[i,k])==list:
                        myargmin[i,k] = myargmin[i,k] + [j]
                    elif myargmin[i,k]==-1:
                        myargmin[i,k] = j
                    else:
                        myargmin[i,k] = [myargmin[i,k]] + [j]

    return myargmin


def df_to_convergence_df(df_in, err_name="errors", time_name="timings", algorithm_name="algorithm", other_names=None, max_time=np.Inf, groups=True, groups_names=None, filters=None):
    """
    Convert a compact Pandas Dataframe with a column ``errors`` containing lists into a long format where these lists are unfolded. This is useful as a post-processor of @run_and_track to be able to easily plot convergence plots with plotly. For instance using @run_and_track to compute and store results,

    >>> @run_and_track(n=[1,2],algorithm_names=["goodalg"]):
    >>> myfun(n=1):
    >>>     costs = [11]
    >>>     time = [0]
    >>>     for i in range(10):
    >>>         # here a dummy cost function over 10 iterations
    >>>         costs.append(10-i)
    >>>         time.append(i)
    >>>     return {"errors": costs, "timings": time}
    >>> # the above code stores the results in a dataframe df
    >>> df = np.load(...)
    >>> # Now we want to plot errors vs time: we use df_to_convergence_df to convert df into a long format for this
    >>> vars = ["n"]
    >>> df_conv = df_to_convergence_df(df, other_names=vars, groups_names=vars)
    >>> # Convergence plots can be easily done with plotly at this stage
    >>> import plotly.express as px
    >>> px.line(df,y="errors",x="timings", facet_col="n") 

    This function will strip all the columns of df which are not error, timings, seed or algorithm_name, and will add an iteration counter. If other columns should be kept, specify their name using the other_names option.

    Parameters
    ----------
    df_in : pandas dataframe
        A dataframe with lists in columns to be unfolded, typically generated by @run_and_track
    err_name : str, optional
        specify a different name of the column in df containing the lists to unfold. Can be useful if you custumized the error names, or if several metrics have been stored in df. By default ``errors``
    time_name : str, optional
        specify a different name of the column in df containing the time lists to unfold. Can be useful if you custumized the timings names. By default ``timings``
    algorithm_name : str, optional
        name of the column containing the algorithm names, by default ``algorithm``
    other_names : list of str, optional
        a list containing the names of columns to keep in the unfolded dataframe, by default None
    max_time : float, optional
        specify a maximum time value after which the dataframe will be truncated, by default np.Inf
    groups : bool, optional
        When plotting line plots with plotly, if other_names have been provided, the lines may loop from end to beginning of the plot. This can be solved by providing a group entry in plotly plots. Groups contains seed by default if seed where provided in df. By default True
    groups_names : list of str, optional
        the name of the columns of df to group, by default None
    filters : dict, optional
        a dictionary with pairs of {column of df: value in that column}, to compute the convergence plot only for the rows in df where the values match the filters values. By default None

    Returns
    -------
    df : pandas dataframe
        A new dataframe in a long format with each row containing a single time, iteration, error value, seed and algorithm_name. Easy to use for plotting convergence plots with plotly.
    """
    # First, we filter undesired rows
    df = df_in.copy() # potentially bad to copy. TODO Better solution
    # sort by algorithm name to avoid issues with lines ordering in plotly
    df = df.sort_values("algorithm")
    df2 = pd.DataFrame()
    # Listify filters
    if filters:
        for key in filters:
            if not type(filters[key])==list:
                filters[key]=[filters[key]]
    # exploring the errors one by one
    for idx_pd,i in enumerate(df[err_name]):
        # filtering
        if filters:
            flag = 0
            for key in filters:
                if df.iloc[idx_pd][key] not in filters[key]:
                    flag=1
                    break
            if flag:
                continue
        
        its = np.arange(0,len(i),1)
        if time_name:
            dic = {
                "it":its,
                time_name: df.iloc[idx_pd][time_name],
                err_name:i,
                algorithm_name:df.iloc[idx_pd][algorithm_name],
                "seed":df.iloc[idx_pd]["seed"]}
        else:
            dic = {
                "it":its,
                err_name:i,
                algorithm_name:df.iloc[idx_pd][algorithm_name],
                "seed":df.iloc[idx_pd]["seed"]}
            # Other custom names to store
        for name in other_names:
            dic.update({name: df.iloc[idx_pd][name]}) 
        df2=pd.concat([df2, pd.DataFrame(dic)], ignore_index=True)

    # cutting time for more regular plots
    if max_time:
        df2 = df2[df2[time_name]<max_time]

    # small preprocessing for grouping plots (avoids lines from end to start in plotly; also more precise, each group represents lines of the same color albeit possibly different runs.)
    if groups:
        zip_arg = []
        if not groups_names:
            # TODO warning
            print("You asked to group convergence plots together, but no parameter has been provided for grouping.")
            groups_names = []
        groups_names += ["seed"]
        for name in groups_names:
            zip_arg.append(df2[name])
        df2["groups"] = list(zip(*zip_arg))
    
    del df
    return df2

def nearest_neighbors_err_at_time_or_it(df, time_stamps=None, it_stamps=None, err_name="errors", time_name="timings"):
    """Adds to dataframe df columns ``err_at_time_xx`` or ``err_at_it_xx`` containing an estimation of errors at given time points or iteration values ``xx``. The estimation is performed for the ``err_at_time_xx`` column using the nearest neighbor error value.

    For a more precise interpolation, check out the ``interpolate_time_and_error`` function.

    Parameters
    ----------
    df : pandas dataframe
        A dataframe containing an ``err_name`` column and a ``time_name`` column 
    time_stamps : list, optional
        a list of values at which the column ``err_name`` should be evaluated, by default None
    it_stamps : list, optional
        a list of indices at which the column ``err_name`` should be evaluated, by default None
    err_name : str, optional
        the name of the error column in df, by default ``errors``
    time_name : str, optional
        the name of the timings columns in df, by default ``timings``

    Returns
    -------
    df: pandas dataframe
        the input dataframe with appended columns ``err_at_time_xx`` or ``err_at_it_xx``. 
    """
    for time in time_stamps:
        store_list = []
        for err,timings in list(zip(df[err_name],df[time_name])):
            # find the error for time closest to the desired time
            # better way since sorted?
            idx_time = np.argmin(np.abs(np.add(timings,-time)))
            err_at_time = err[idx_time]
            store_list.append(err_at_time)
        df["err_at_time_"+str(time)] = store_list
    for it in it_stamps:
        store_list = []
        for err in df[err_name]:
            err_at_it = err[it]
            store_list.append(err_at_it)
        df["err_at_it_"+str(it)] = store_list
    return df
            
def regroup_columns(df,keys=None, how_many=None, textify=True):
    """
    Because we split input lists in DataFrames for storage, it may be convenient to re-introduce the original lists as columns on user demand.

    For instance, to grid over a parameter r=[r1,r2,r3] which is naturally represented as a list, in the current version of the shootout toolbox, we are forced to store the results in three columns labeled "r_1", "r_2", "r_3".

    Parameters
    ----------
    df: pandas Dataframe
        a dataframe containing columns to regroup
    keys: list of strings, or string 
        a list of column names of the form foo_n where foo is in keys and n is an integer.
    how_many: int
        the number of elements that were splitted as dataframe columns
    textify: bool, optional
        makes strings inputs instead of list of integers because this makes life 8 times easier with Pandas and Plotly, default True
    """
    if type(keys)!=list:
        keys = [keys]
    for name in keys:
        df[name] = pd.Series([[] for i in range(len(df))])
        for j in range(len(df)):
            if textify:
                df[name][j] = str([df[name+"_"+str(i)][j] for i in range(how_many)])
            else: 
                df[name][j] = [df[name+"_"+str(i)][j] for i in range(how_many)]
        # then we remove the columns that were merged
        df = df.drop([name+"_"+str(i) for i in range(how_many)], axis=1)    
    return df

def interpolate_time_and_error(df, err_name="errors", time_name="timings", k=0, logtime=False, npoints=500, adaptive_grid=False, alg_name="algorithm"):
    """
    Interpolates several error curves so that they all lie on the same grid. There is one grid per algorithm name by default. To use with caution as this can significantly bend the error curves.

    Parameters
    ----------
    df : pandas dataframe
        raw shootout results dataframe with errors and time in lists.
    k  : int, default 0
        the max time for interpolation will be the (longest - k) runtime for the algorithms. Set to k>0 if k runs are abnormally long.
    logtime : boolean, default False
        choose if interpolation grid is linear (False) or logarithmic (True). Set to True when timings are very different between several runs.
    npoints : int, default 500
        number of iterpolation points.
    adaptive_grid : bool, default False
        determines if each test condition has its own time grid. If True, a dirty BAD hack is used: shootout runs samples in most intern loop, therefore we can compute grids for each block of rows in df cut according to periodicity of seed.
    alg_name : string, default "algorithm"
        the string value of the key in df that contains the algorithm name.
    """
    # Check if grid is individual or global
    if adaptive_grid:
        # empty columns for init
        df["errors_interp"] = df[err_name]
        df["timings_interp"] = df[time_name]
        # We work on blocks of df cut according to seed periodicity
        if alg_name in df.keys():
            nb_algs = np.maximum(len(np.unique(df["algorithm"])),1) #0 not allowed, in case algorithms are not provided
        else:
            nb_algs = 1
        seed_periodicity = (df["seed"].max() + 1)*nb_algs
        nb_rows = len(df)
        nb_blocks = int(nb_rows/seed_periodicity)
        for block_idx in range(nb_blocks):
            for alg_idx in range(nb_algs):
                maxtime_list = []
                for row in df[time_name][seed_periodicity*block_idx+alg_idx:seed_periodicity*(block_idx+1):nb_algs]: # todo check indices
                    maxtime_list.append(row[-1])
                maxtime = np.sort(maxtime_list)[-1]
                # then we create a grid
                if logtime:
                    time_grid = np.logspace(0, np.log10(maxtime), npoints)
                else:
                    time_grid = np.linspace(0, maxtime, npoints)
                # now we populate the time column for the block and interpolate
                for idx in range(seed_periodicity*block_idx+alg_idx, seed_periodicity*(block_idx+1), nb_algs):
                    df["timings_interp"][idx] = time_grid
                    new_errors = np.interp(time_grid, df[time_name][idx], df[err_name][idx])
                    df["errors_interp"][idx]=new_errors

    else:
        # First we look for the k-th max timing over all runs
        maxtime_list = []
        for timings in df[time_name]:
            maxtime_list.append(timings[-1])
        kmaxtime = np.sort(maxtime_list)[-1-k]
        # then we create a grid based on that time
        if logtime:
            time_grid = np.logspace(0, np.log10(kmaxtime), npoints)
        else:
            time_grid = np.linspace(0, kmaxtime, npoints)
        # creating dummy columns
        df["errors_interp"] = df[err_name]
        df["timings_interp"] = df[time_name]
        # now we interpolate each error on that grid
        for idx_errors, errors in enumerate(df[err_name]):
            new_errors = np.interp(time_grid, df[time_name][idx_errors], errors)
            # we can update the dataframe on the fly
            df["errors_interp"][idx_errors]=new_errors
            df["timings_interp"][idx_errors]=time_grid

    return df


def median_convergence_plot(df_conv, type_x="iterations", err_name="errors", time_name="timings", mean=False):
    """
    Computes the mean or median of several converge plots with respect to either time or iterations (this is determined by ``type_x``). By mean/median of several curve, we mean the mean/median at each ``type_x`` value.

    Parameters
    ----------
    df : pandas dataframe
        input dataframe with error at each iteration split in rows
        requires timings to be aligned by linear interpolation first.
    type_x : string, default ``timings``
        choose if the median is over time or over iterations.
    means : boolean, default False
        if True, compute the mean of the curves, otherwise the median is computed.
    """
    # TODO: use another syntax more similar to above?
    # we use the groupby function; we will groupy by everything except:
    # - errors (we want to median them)
    # - seeds (they don't matter since we use conditional mean over everything else)
    # - timings
    df = df_conv.copy()
    df.pop("seed")
    #if type == "timings":
        ## we store time to put it back at the end
        #timings_saved = df[time_name] 
    if type_x=="iterations":
        df.pop(time_name) # we always pop time, since we made sure it is aligned with iterations
    elif type_x=="timings":
        df.pop("it")
    # else pop custom
    else:
        df.pop(type_x)

    # iterations behave like an index for computing the median
    try:
        df.pop("groups") # good idea?
    except:
        pass

    namelist = list(df.keys())
    namelist.remove(err_name)

    #print(namelist)

    if mean:
        df_med = df.groupby(namelist, as_index=False).mean() 
    else:
        df_med = df.groupby(namelist, as_index=False).median() 
    df_02 = df.groupby(namelist, as_index=False).quantile(q=0.2) 
    df_08 = df.groupby(namelist, as_index=False).quantile(q=0.8) 
    df_med["q_errors_p"] = df_08[err_name]-df_med[err_name]
    df_med["q_errors_m"] = df_med[err_name]-df_02[err_name]

    #if type=="timings":
        #df_med[time_name] = timings_saved

    return df_med#, df_time.groupby(namelist_time).median()


# A first test for my shoot-out decorators
import itertools
import copy
import os
import pandas
from datetime import datetime
from shootout.methods.post_processors import regroup_columns

def run_and_track(add_track=None, algorithm_names=None, path_store=None, name_store=None,
                    verbose=True, skip=False, **kwa):
    """ 
    This function is the main ingredient of shootout. It is a decorator, to run a python function ``my_script`` with a set of inputs to grid on. The outputs are stored in a pandas DataFrame which is stored locally.
    
    Shootout is designed to make optimization algorithms comparison easier, so the target application is to lift up a code comparing algorithms on single parameter settings to a full comparison and storing all parameters and results.

    A few restrictions are placed on the wrapped function:

        - outputs is one dictionary
        - the errors, timings and any other algorithm dependent quantity must be inside lists of length the number of algorithms.

    In single_method mode, 2. is not applied. We allow the user to simply provide anything that will be added as an object to the df.

    A toy example:

    >>> @run_and_track(nb_seeds=5, n=[5,10], m=[4,8])
    >>> def my_script(n=7, m=9, p=12):
    >>>     return { "something": n+m+p }

    will return a dataframe containing four columns ``n`` , ``m`` , ``something`` and ``seed`` , with 5x2x2 rows (5 runs which are here exactly the same, two values for n and two values for m).    
    Note that the function should output a dictionary.

    **Syntax:**
    In the case of a single algorithm to run with several hyperparameters, the syntax looks like this:
    
    >>> @run_and_track(algorithm_names=['My favorite Algorithm'], hyperparameter=[0.1,1,10], seed=[1,15,975])
    >>> def my_script(n=10, m=10, hyperparameter=0, seed=0):
    >>>     # some random generation using seed, e.g. with numpy randomstate
    >>>     # seed will be populated by numbers in the seed list
    >>>     # and hyperparameter with 0.1, 1 and 10
    >>>     errors, timings, anything_else = my_algorithm(random_data,n,m,hyperparameter)
    >>>     return {"errors": errors, "timings": timings}
    
    where errors and timings must have the same length.
    
    For several algorithms, say two, the syntax for the returns of ``my_script`` function is similar but slightly different:
    
    >>> @run_and_track(algorithm_names=['My Favorite Algorithm', 'my annoying competitor'], seed=[1,17,597], hyperparameter=[0.1,1,10])
    >>> def my_script(n=10, m=10, hyperparameter=0, seed=0):
    >>>     # some random generation using seed, e.g. with numpy randomstate
    >>>     # seed will be populated by numbers from 0 to nb_seeds
    >>>     # and hyperparameter with 0.1, 1 and 10
    >>>     errors, timings, anything_else = my_algorithm(random_data,n,m,hyperparameter)
    >>>     errors2, timings2, anything_else2 = my_competitor(random_data,n,m,hyperparameter)
    >>>     return {"errors": [errors, errors2], "timings": [timings, timings2]}

    The naming of the inputs and outputs is completely free. However, naming errors/loss/utility collected in a list exactly ``errors`` and ``timings`` (time starting from 0) will allow for immediate processing with utilities from shootout such as plotting convergence curves.

    The recommended syntax when many parameters are involved is to define all the parameters in a dictionary, see ``**kwa`` below. For the signature of the inner function, one should also use a dictionary that makes use of the same keys.

    Parameters
    ----------
    **kwa: dictionary, optional
        A dictionary containing all the elements in the signature of the function to run. Typically use in the following fashion:

        >>> vars = dict(
        >>>     {
        >>>         "n": [10,15],
        >>>         "m": [25,30],
        >>>         "noise": 0.25,
        >>>         "r": 3,
        >>>         "seed": [1,2,3,4]
        >>>     }
        >>> )
        >>> @run_and_track(**vars)
        >>> def myalgorithm(**v):
        >>>     xxx = v["n"] + ...
        
        Another syntax is tolerated although not recommended, where the user may mismatch the signatures of "@run_and_track" and the ran function as such:

        >>> @run_and_track(n=[10,15], m=[25,30], noise=0.25, add_track={"mytrack":[7,12,20]}, name_store="test-df", nb_seeds=2, algorithm_names=["method1"])
        >>> def myalgorithm(n=10,m=20,r=3,random=True,noise=0.5):
        >>>     ...

        the entries in run_and_track must then be repeated in the function to be run. Note that both methods cannot be combined, so as to promote the use of the first method which is cleaner (all function parameters used for the run will be stored and the variables are not duplicated).
        By default None.
    add_track : dictionary, optional
        a dictionary of values to store manually. The inputs and outputs are stored anyway so this option should be avoided if possible. By default None
    algorithm_names : list, optional
        provide a list of name for each algorithm that will be run in the script. The names should match the outputs of the script if the outputs are lists (same size and order). Using this input is recommended for easily naming plots in post-processing stages. By default None
    path_store : string, optional
        the path (relative or absolute) to the directory where the DataFrame will be stored, by default None, and default directory is the current working directory.
    name_store : string, optional
        name of the file containing the DataFrame, by default None and the default name is the date in long format.
    verbose : bool, optional
        choose wether run_and_track prints its progress, by default True
    skip : bool, optional
        Set to True to skip computations, by default False.
    """
    #TODO: better error messages/error catching \
    #    - missmatches input names / loop variables
    #    - solo fun vs several
    #    - algorithm_names length does not match outputs
    
    def inner_run_and_track(fun):    
        print(f"verbose works?? {verbose}")
        # Preprocessing things
        if skip:
            return

        # Preprocessing: converting any single value in kwa to a singleton list
        # also tracking names of list parameters for rebuilding columns
        list_params_keys = []
        list_params_lengths = []
        for i in kwa:
            if type(kwa[i])!= list:
                #print("Converting single parameter swipe to singleton")
                kwa[i] = [kwa[i]]
            else:
                # get length of items in the list
                if type(kwa[i][0])==list:
                    list_params_keys.append(i)
                    list_params_lengths.append(len(kwa[i][0]))

        if algorithm_names:
            if len(algorithm_names)>1:
                single_method=False
            else:
                single_method=True
        else:
            # TODO Warning
            print("Consider adding algorithm names to benefit from automatic labeling of runs")

        # Before all, initialize our storage DataFrame
        df = pandas.DataFrame()

        # Now we enter the big loop
        for params in itertools.product(*[kwa[key] for key in kwa]):
            # Discrete printing for following the run
            if verbose:
                print(f"Params values are currently:")
                for i,key in enumerate(kwa):
                    print("   ", key+":", params[i])
            # params are not a dictionary so we make function read them by building a dictionary
            dic = kwa
            for i,elem in enumerate(kwa.keys()):
                dic[elem] = params[i]
            # Calling our script with the correct hyperparameters, everything else is defaults
            outputs = fun(**dic)

            # Storing all results and parameters in a dataframe
            # This is inputs, outputs (both fixed and variable)
            store_dic = {}
            store_dic.update(dic)
            # Optional manual tracking
            if add_track:
                store_dic.update(add_track)
            # A counterintuitive step is to unbracket all the singleton list parameter values in the dictionary, but split other lists. Indeed, pandas will have to consider these values as objects to store several lines simultaneously, and this does the job efficiently.
            dic_copy = copy.copy(store_dic)
            # copy to avoid changing the dictionary online, which fails
            for elem in dic_copy:
                if type(store_dic[elem]) == list:
                    if len(store_dic[elem])==1:
                        store_dic[elem] = store_dic[elem][0]
                    else:
                        # we go over all elements of the list and split them in the dictionary TODO improve
                        for i in range(len(store_dic[elem])):
                            store_dic[elem+"_"+str(i)] = store_dic[elem][i]
                        # Then remove the original list
                        # Otherwise when creating the DataFrame, we have lists of various sizes in the dic and this is not accepted as input
                        del store_dic[elem]
            del dic_copy
            if algorithm_names:
                store_dic.update({"algorithm": algorithm_names})

            if single_method:
                for elem in outputs:
                    outputs[elem] = [outputs[elem]] # double bracket makes pandas objectify this
            store_dic.update(outputs)

            # We can finally update the pandas DataFrame with all the run information
            df_temp = pandas.DataFrame(store_dic)
            df= pandas.concat([df,df_temp], ignore_index=True)

        # update dataframe to reconstruct the input lists as true lists
        for i in range(len(list_params_keys)):
            df = regroup_columns(df, keys=list_params_keys[i], how_many=list_params_lengths[i])

        # To store the dataframe, we use the provided path, otherwise we place at runpath
        if path_store:
            path_store_full = path_store
        else:
            path_store_full = os.getcwd()+"/"
        if name_store:
            path_store_full += name_store
        else:
            path_store_full += "run-{}".format(datetime.today().strftime('%Y-%m-%d_%H-%M'))
        df.to_pickle(path_store_full)
        return fun # This is what the wrapped function will actually be
    return inner_run_and_track


from shootout.methods.post_processors import nearest_neighbors_err_at_time_or_it, df_to_convergence_df, find_best_at_all_thresh, regroup_columns, interpolate_time_and_error, median_convergence_plot, my_argmin
import pandas as pd
import numpy as np
import os

path = os.path.dirname(__file__)
df = pd.read_pickle(path+"/run-example")

# TODO: myargmin
def test_my_argmin():
    a = np.array([[[1,1],[1,4]],[[5,6],[6,6]]])
    out = my_argmin(a)
    assert out[0,0] == [0,1]
    assert out[0,1] == 0
    assert out[1,0] == 0
    assert out[1,1] == [0,1]

# TODO: check that all args are tested

def test_find_best_at_all_thresh(df=df):
    thresh = np.logspace(5,2,50) 
    scores_time, scores_it, timings, iterations = find_best_at_all_thresh(df,thresh, 2)
    # stupid unit test
    assert scores_time[0,0] == 35

# Testing df_to_convergence_df
def test_df_to_convergence_df(df=df):
    other_names = ["delta"]  #useless, just to invoke the option
    df2 = df_to_convergence_df(df, other_names=other_names, max_time=1.5,
                                filters={"seed":[0,1], "loss":"l2"}, groups=True, groups_names=other_names)
    # another stupid test
    assert len(df2) == 410

# Testing err_at_time_or_it
def test_nearest_neighbors_err_at_time_or_it(df=df):
    df = nearest_neighbors_err_at_time_or_it(df, time_stamps=[0.1,0.5,1], it_stamps=[0,10,40])
    assert df["err_at_time_1"][0]<1.132584e-7 and df["err_at_time_1"][0]>1.132582e-7

# Testing regrouping
# manually adding columns
def test_regroup_columns(df=df):
    df["ranks_0"]=0
    df["ranks_1"]=0
    df = regroup_columns(df,keys=["ranks"], how_many=2, textify=True)
    # output is a string
    assert df["ranks"][0]=='[0, 0]'

def test_interpolate_time_and_error(df=df):
    df = interpolate_time_and_error(df, logtime=True, npoints=39, adaptive_grid=True)
    assert  32 > df["errors_interp"][69][0] > 31.8

def test_median_convergence_plot(df=df):
    df = interpolate_time_and_error(df, logtime=True, npoints=39, adaptive_grid=True)
    df_l2_conv = df_to_convergence_df(df, groups=True, groups_names=[], other_names=[],
                                filters={"loss":"l2"}, err_name="errors_interp", time_name="timings_interp")
    df_l2_conv = df_l2_conv.rename(columns={"timings_interp": "timings", "errors_interp": "errors"})
    df_l2_conv_it = df_to_convergence_df(df, groups=True, groups_names=[], other_names=[],
                                filters={"loss":"l2"})
    df_l2_conv_median_time = median_convergence_plot(df_l2_conv, type_x="timings")
    df_l2_conv_median_it = median_convergence_plot(df_l2_conv_it, type_x="iterations")
    assert 1.15e-8<df_l2_conv_median_it["q_errors_p"][0]<1.17e-8 and 1.4e-10<df_l2_conv_median_time["q_errors_p"][0]<1.41e-10


from shootout.methods.runners import run_and_track
import pandas as pd
import os
import numpy as np

# Testing run_and_track
# check that a DataFrame is created and that this function runs smoothly
def test_run_and_track_multiple_algs():  
    vars = dict({
        "n": [10,15],
        "m": [25,30],
        "listpar":[[1,2,3],[4,5,6]],
        "noise": 0.25,
        "r": 3,
        "seed": [1,2]
    })
    @run_and_track(add_track={"mytrack":[7,12,20]}, name_store="test-df", algorithm_names=["method1","method2"], **vars)
    def myalgorithm(**v):
        w = v["m"]+v["n"]+v["r"]+v["noise"]+v["seed"]
        h = [1,2,3,4]
        err = [v["n"]-v["m"], v["n"]+v["m"]+v["seed"]]
        time = [0,0.2]
        return {"errors":[err,err],"timings": [time,time],"myoutput":[h,h],"scalaroutput":7}
    # We check if the dataframe has been created by loading it
    df = pd.read_pickle("test-df")
    # picking a random element to check
    assert df["errors"][6][1] == 37
    # deleting temporary df
    os.remove("test-df")
    #return df


# Another test for single algorithm
def test_run_and_track_single_algorithm():
    vars = dict(
        {
            "n": [10,15],
            "m": [25,30],
            "noise": 0.25,
            "r": 3,
            "random": True
        }
    )
    @run_and_track(add_track={"mytrack":[7,12,20]}, name_store="test-df", algorithm_names=["method1"], **vars)
    def myalgorithm(**v):
        w = v['m']+v['n']+v['r']+v['noise']
        h = np.array([[1,2,3,4],[5,6,7,8]])
        err = [v['n']-v['m'], v['n']+v['m'], v['n']**2]
        time = [0,0.2 ]
        return {"errors":err,"timings": time, "myoutput":h, "scalaroutput": 7}
    # We check if the dataframe has been created by loading it
    df = pd.read_pickle("test-df")
    # picking a random element to check
    print(df)
    assert df["errors"][0][0] == -15
    # deleting temporary df
    os.remove("test-df")